// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file MatVecMultipler.h */

#ifndef AMDIS_MATVECMULTIPLIER_H
#define AMDIS_MATVECMULTIPLIER_H

#include<vector>

#include "Global.h"
#include "MemoryManager.h"
//#include "Boundary.h"
#include "DOFVector.h"
#include "DOFMatrix.h"
#include "MatrixVector.h"
#include "time.h"

namespace AMDiS {

  template<typename T> class DOFVector;

  // ===========================================================================
  // ===== class MatVecMultiplier ==============================================
  // ===========================================================================

  /** 
   * \ingroup Solver
   *
   * \brief
   * Interface for matrix-vector multiplication.
   */
  template<typename Vector>
  class MatVecMultiplier
  {
  public:
    MEMORY_MANAGED(MatVecMultiplier);

    /** \brief
     * Destructor.
     */
    virtual ~MatVecMultiplier() {};

    /** \brief
     * To be overloaded by concrete implementations.
     */
    virtual void matVec(MatrixTranspose transpose,
			const Vector    &vec, 
			Vector          &result,
			bool add = false) = 0;
  };

  // ===========================================================================
  // ===== class StandardMatVec ================================================
  // ===========================================================================

  /** 
   * \ingroup Solver
   *
   * \brief
   * Implementation of the standard matrix vector multiplication.
   */
  template<typename MatrixType, typename VectorType>
  class StandardMatVec : public MatVecMultiplier<VectorType>
  {
  public:
    MEMORY_MANAGED(StandardMatVec<MatrixType COMMA VectorType>);

    /** \brief
     * 
     */
    StandardMatVec(MatrixType *mat) 
      : matrix(mat)
    {};

    virtual ~StandardMatVec() {};

    /** \brief
     * Implements \ref MatVecMultiplier::matVec().
     */
    virtual void matVec(MatrixTranspose   transpose,
			const VectorType &x, 
			VectorType       &result,
			bool add = false)  {
      FUNCNAME("StandardMatVec::matVec()");

      WARNING("Non specialised function should not be called\n");
    };

    /** \brief
     * Sets \ref matrix.
     */
    virtual void setMatrix(MatrixType *m) { 
      matrix = m; 
    };

    virtual MatrixType* getMatrix() { 
      return matrix; 
    };

    virtual const MatrixType* getMatrix() const { 
      return matrix; 
    };

  protected:
    /** \brief
     * Matrix used for matrix-vector multiplication.
     */
    MatrixType *matrix;
  };



  template<typename MatrixType, typename VectorType>
  class MassLumpingMatVec : public StandardMatVec<MatrixType, VectorType>
  {
  public:
    /**
     * Empty constructor.
     */
    MassLumpingMatVec()
      : StandardMatVec<MatrixType, VectorType>(NULL)
    {};

    /** \brief
     * 
     */
    MassLumpingMatVec(MatrixType *mat)
      : StandardMatVec<MatrixType, VectorType>(NULL)
    {
      calcMassLumpingMatrix(mat);
    };

    virtual ~MassLumpingMatVec() {
      if (this->matrix) 
	DELETE this->matrix;
    };

    /** \brief
     * Sets \ref matrix.
     */
    void setMatrix(MatrixType *m) {
      FUNCNAME("MassLumpingMatVec::setMatrix()");

      ERROR_EXIT("should not be called\n");
    };

    void calcMassLumpingMatrix(MatrixType *m);
  };



  template<> void 
  StandardMatVec<DOFMatrix, DOFVector<double> >::matVec(MatrixTranspose transpose,
							const DOFVector<double> &x,
							DOFVector<double> &result,
							bool add);


  class SystemVector;

  template <> void 
  StandardMatVec<Matrix<DOFMatrix*>, SystemVector>::matVec(
							   MatrixTranspose transpose,
							   const SystemVector &x, 
							   SystemVector &result,
							   bool add
							   );


}
#endif // AMDIS_MATVECMULTIPLIER_H
