// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file RecoveryEstimator.h */

#ifndef AMDIS_MYRECOVERYESTIMATOR_H
#define AMDIS_MYRECOVERYESTIMATOR_H

#include "Estimator.h"
#include "Recovery.h"

namespace AMDiS {

  // ============================================================================
  // ===== class RecoveryEstimator ============================================
  // ============================================================================

  /** \brief
   * Error estimator using the recovery gradient of the finite element solution.
   */
  class RecoveryEstimator : public Estimator
  {
  public:
    MEMORY_MANAGED(RecoveryEstimator);

    /** \brief
     * Creator class.
     */
    class Creator : public EstimatorCreator
    {
    public:
      MEMORY_MANAGED(Creator);

      Creator() : EstimatorCreator(), uh(NULL) {};

      virtual ~Creator() {};

      inline void setSolution(DOFVector<double> *uh_)
      {
	uh = uh_;
      };

      /** \brief
       * Returns a new Estimator object.
       */
      Estimator* create()
      {
	return NEW RecoveryEstimator(name, uh, row);
      };

    protected:
      DOFVector<double> *uh;
    };

    /** \brief
     * constructor
     */
    RecoveryEstimator(::std::string name, DOFVector<double> *uh, int r = -1);

    /** \brief
     * destructor.
     */
    virtual ~RecoveryEstimator() {};

    /** \brief
     * implements \ref Estimator::estimate().
     */
    virtual double estimate(double timestep = 0.0);


    /*----- Setting functions -----*/

    // Sets uh.
    inline void setUh(DOFVector<double> *uh)
    {
      uh_ = uh;
    };

    // Sets f.
    inline void setFct(AbstractFunction<double, WorldVector<double> > *fct)
    {
      f_vec = fct;
    };

    inline void setFct(AbstractFunction<double, double> *fct)
    {
      f_scal = fct;
    };

    // Sets auxiliar vector.
    inline void setAuxVec(DOFVector<double> *uh)
    {
      aux_vec = uh;
    };


    /*----- Getting functions -----*/

    // Gets recovery gradient.
    inline DOFVector<WorldVector<double> >* getRecGrd()
    {
      return rec_grd;
    };

    // Gets higher-order recovery solution.
    inline DOFVector<double>* getRecUh()
    {
      return rec_uh;
    };


  protected:     // protected members

    /** \brief
     * finite element solution
     */
    DOFVector<double> *uh_;

    /** \brief
     * absolute or relative error?
     */
    int relative_;

    /** \brief
     * constant for scaling the estimator
     */
    double C;

    /** \brief
     * recovery method
     */
    int method;

    /** \brief
     * Working finite element space
     */
    const FiniteElemSpace *feSpace;

    /** \brief
     * Degree of corresponding basic functions
     */
    int degree;

    /** \brief
     * Basis functions for recovery vector.
     */
    const BasisFunction *rec_basFcts;

    /** \brief
     * Recovery gradient
     */
    DOFVector<WorldVector<double> > *rec_grd;

    /** \brief
     * Higher-order recovery solution
     */
    DOFVector<double> *rec_uh;

    /** \brief
     * Diffusion coefficient (for flux recovery)
     */
    AbstractFunction<double, WorldVector<double> > *f_vec;
    AbstractFunction<double, double> *f_scal;

    /** \brief
     * auxiliar vector
     */
    DOFVector<double> *aux_vec;

    /** \brief
     * Recovery structure.
     */
    Recovery *rec_struct;
  };

}

#endif
