#include <stdio.h>
#include <string>
#include <fstream>
#include <cmath>

#include "VtkWriter.h"
#include "DataCollector.h"
#include "DOFVector.h"
#include "SurfaceRegion_ED.h"
#include "ElementRegion_ED.h"

namespace AMDiS { 

  int VtkWriter::writeFile(const char *name)
  {
    int nVertices = (*dc_)[0]->getNumberVertices();
    int nElements = (*dc_)[0]->getNumberElements();
    int vertices = (*dc_)[0]->getMesh()->getGeo(VERTEX);

    if ((dim_ == 2) && (degree_ == 2)) {
      nVertices += (*dc_)[0]->getNumberInterpPoints();
      nElements *= 4;
    } else if ((dim_ == 2) && (degree_ == 3)) {
      nVertices += (*dc_)[0]->getNumberInterpPoints();
      nElements *= 9;
    } else if ((dim_ == 2) && (degree_ == 4)) {
      nVertices += (*dc_)[0]->getNumberInterpPoints();
      nElements *= 16;
    }
    
    ::std::ofstream file;
    file.open(name);
    
    file << "<?xml version=\"1.0\"?>" << ::std::endl;
    file << "<VTKFile type=\"UnstructuredGrid\" version=\"0.1\" byte_order=\"LittleEndian\">" << ::std::endl;
    file << "  <UnstructuredGrid>" << ::std::endl;
    file << "    <Piece NumberOfPoints=\"" << nVertices 
	 << "\" NumberOfCells=\"" << nElements << "\">" << ::std::endl;
    file << "      <Points>" << ::std::endl;
    file << "        <DataArray type=\"Float32\" NumberOfComponents=\"3\" format=\"ascii\">" << ::std::endl;
    
    writeVertexCoords(file);
    
    file << "        </DataArray>" << ::std::endl;
    file << "      </Points>" << ::std::endl;
    file << "      <Cells>" << ::std::endl;
    file << "        <DataArray type=\"Int32\" Name=\"offsets\">" << ::std::endl;
    
    for (int i = 0; i < nElements; i++) {
      file << " " << (i + 1) * vertices << ::std::endl;
    }

    file << "        </DataArray>" << ::std::endl;
    file << "        <DataArray type=\"UInt8\" Name=\"types\">" << ::std::endl;

    for (int i = 0; i < nElements; i++) {
      switch (vertices) {
      case 2:
	file << " 3" << ::std::endl;
	break;
      case 3: 
	file << " 5" << ::std::endl;
	break;
      case 4:
	file << " 10" << ::std::endl;
	break;		    
      default:
	break;
      }	   
    }

    file << "        </DataArray>" << ::std::endl;
    file << "        <DataArray type=\"Int32\" Name=\"connectivity\">" << ::std::endl;
    
    writeConnectivity(file);
    
    file << "        </DataArray>" << ::std::endl;    
    file << "      </Cells>" << ::std::endl;
    file << "      <PointData>" << ::std::endl;
    
    for (int i = 0; i < static_cast<int>(dc_->size()); i++) {
      file << "        <DataArray type=\"Float32\" Name=\"value" << i 
	   << "\" format=\"ascii\">" << ::std::endl;
      
      writeVertexValues(file, i);
      
      file << "        </DataArray>" << ::std::endl;
    }
    
    file << "      </PointData>" << ::std::endl;
    file << "    </Piece>" << ::std::endl;
    file << "  </UnstructuredGrid>" << ::std::endl;
    file << "</VTKFile>" << ::std::endl;
    
    
    file.close();
    
    return 0;
  }


  void VtkWriter::writeVertexCoords(::std::ofstream &file)
  {
    DOFVector< ::std::list<VertexInfo> > *vertexInfos = (*dc_)[0]->getVertexInfos();
    DOFVector< ::std::list<VertexInfo> >::Iterator it(vertexInfos, USED_DOFS);
    int counter = 0;

    // For all DOFs of vertices, write the coordinates.
    for (it.reset(); !it.end(); ++it) {
      // for all vertex infos of this DOF
      ::std::list<VertexInfo>::iterator it2;
      for (it2 = it->begin(); it2 != it->end(); ++it2) {
	it2->outputIndex = counter++;
	writeCoord(file, it2->coords);
      }
    }

    // For the second dim case, write also the interpolation points.
    if ((dim_ == 2) && (degree_ > 1)) {
      DOFVector< ::std::list< WorldVector<double> > > *interpPointCoords = (*dc_)[0]->getInterpPointCoords();
      DOFVector< ::std::list< WorldVector<double> > >::Iterator pointIt(interpPointCoords, USED_DOFS);
      
      for (pointIt.reset(); !pointIt.end(); ++pointIt) {
	::std::list< WorldVector<double> >::iterator it2;
	for (it2 = pointIt->begin(); it2 != pointIt->end(); ++it2) {
	  writeCoord(file, *it2);
	}
      }
    }
  }


  void VtkWriter::writeVertexValues(::std::ofstream &file, int componentNo)
  { 
    DOFVector<int> *interpPointInd;
    DOFVector<double> *values;
    DOFVector< ::std::list<WorldVector<double> > > *dofCoords;

#ifdef _OPENMP
#pragma omp critical 
#endif
    {   
      interpPointInd = (*dc_)[componentNo]->getInterpPointInd();
      values = (*dc_)[componentNo]->getValues();
      dofCoords = (*dc_)[componentNo]->getDofCoords();
    }
    
    DOFVector<int>::Iterator intPointIt(interpPointInd, USED_DOFS);
    DOFVector<double>::Iterator valueIt(values, USED_DOFS);
    DOFVector< ::std::list<WorldVector<double> > >::Iterator coordIt(dofCoords, USED_DOFS);
          
    // Write the values for all vertex DOFs.
    for (intPointIt.reset(), valueIt.reset(), coordIt.reset();
	 !intPointIt.end(); 
	 ++intPointIt, ++valueIt, ++coordIt) {

      if (*intPointIt == -2) {
	for (int i = 0; i < static_cast<int>(coordIt->size()); i++) {
	  file << " " << (fabs(*valueIt) < 1e-40 ? 0.0 : *valueIt) << ::std::endl;
	}
      }
    }        

    // For the second dim case, write also the values of the interpolation points.
    if ((dim_ == 2) && (degree_ > 1)) {
      DOFVector< ::std::list<WorldVector<double> > >::Iterator 
	interpCoordIt((*dc_)[componentNo]->getInterpPointCoords(), USED_DOFS);
    
      for (intPointIt.reset(), valueIt.reset(), interpCoordIt.reset();
	   !intPointIt.end(); 
	   ++intPointIt, ++valueIt, ++interpCoordIt) {      
	
	if (*intPointIt >= 0) {
	  for (int i = 0; i < static_cast<int>(interpCoordIt->size()); i++) {
	    file << " " << (fabs(*valueIt) < 1e-40 ? 0.0 : *valueIt) << ::std::endl;
	  }
	}
      }
    }    
  }


  void VtkWriter::writeConnectivity(::std::ofstream &file) 
  {
    // For the second dim case, and if higher order Lagrange elements are used,
    // write the connectivity by extra functions.
    if ((dim_ == 2) && (degree_ == 2)) {
      writeConnectivity_dim2_degree2(file);
    } else if ((dim_ == 2) && (degree_ == 3)) {
      writeConnectivity_dim2_degree3(file);
    } else if ((dim_ == 2) && (degree_ == 4)) {
      writeConnectivity_dim2_degree4(file);   
    } else {
      ::std::list<ElementInfo> *elements = (*dc_)[0]->getElementInfos();
      ::std::list<ElementInfo>::iterator elementIt;
      int vertices = (*dc_)[0]->getMesh()->getGeo(VERTEX);
      
      for (elementIt = elements->begin(); elementIt != elements->end(); ++elementIt) {
	// for all vertices
	for (int i = 0; i < vertices; i++) {
	  file << " " << elementIt->vertexInfo[i]->outputIndex;
	}
	file << ::std::endl;
      } 
    }
  }


  void VtkWriter::writeConnectivity_dim2_degree2(::std::ofstream &file)
  {  
    ::std::list<ElementInfo> *elements = (*dc_)[0]->getElementInfos();
    ::std::list<ElementInfo>::iterator elementIt;

    ::std::vector< ::std::vector<int> > *interpPoints = (*dc_)[0]->getInterpPoints();
    ::std::vector< ::std::vector<int> >::iterator pointIt;

    int nVertices = (*dc_)[0]->getNumberVertices();

    for (pointIt = interpPoints->begin(), elementIt = elements->begin(); 
	 pointIt != interpPoints->end(); 
	 ++pointIt, ++elementIt) {
    
      file << " " << elementIt->vertexInfo[0]->outputIndex
	   << " " << (*pointIt)[1] + nVertices
	   << " " << (*pointIt)[2] + nVertices << ::std::endl;

      file << " " << elementIt->vertexInfo[2]->outputIndex 
	   << " " << (*pointIt)[0] + nVertices
	   << " " << (*pointIt)[1] + nVertices << ::std::endl;

      file << " " << elementIt->vertexInfo[1]->outputIndex
	   << " " << (*pointIt)[0] + nVertices
	   << " " << (*pointIt)[2] + nVertices << ::std::endl;

      file << " " << (*pointIt)[0] + nVertices
	   << " " << (*pointIt)[1] + nVertices 
	   << " " << (*pointIt)[2] + nVertices << ::std::endl;
    }

  }


  void VtkWriter::writeConnectivity_dim2_degree3(::std::ofstream &file)
  {  
    ::std::list<ElementInfo> *elements = (*dc_)[0]->getElementInfos();
    ::std::list<ElementInfo>::iterator elementIt;

    ::std::vector< ::std::vector<int> > *interpPoints = (*dc_)[0]->getInterpPoints();
    ::std::vector< ::std::vector<int> >::iterator pointIt;

    int nVertices = (*dc_)[0]->getNumberVertices();

    for (pointIt = interpPoints->begin(), elementIt = elements->begin(); 
	 pointIt != interpPoints->end(); 
	 ++pointIt, ++elementIt) {
                          
      file << " " << elementIt->vertexInfo[0]->outputIndex
	   << " " << (*pointIt)[3] + nVertices
	   << " " << (*pointIt)[4] + nVertices << ::std::endl;

      file << " " << (*pointIt)[4] + nVertices
	   << " " << (*pointIt)[5] + nVertices
	   << " " << (*pointIt)[6] + nVertices << ::std::endl;

      file << " " << (*pointIt)[3] + nVertices
	   << " " << (*pointIt)[4] + nVertices
	   << " " << (*pointIt)[6] + nVertices << ::std::endl;

      file << " " << (*pointIt)[2] + nVertices
	   << " " << (*pointIt)[3] + nVertices
	   << " " << (*pointIt)[6] + nVertices << ::std::endl;

      file << " " << elementIt->vertexInfo[1]->outputIndex
	   << " " << (*pointIt)[0] + nVertices
	   << " " << (*pointIt)[5] + nVertices << ::std::endl;

      file << " " << (*pointIt)[0] + nVertices
	   << " " << (*pointIt)[6] + nVertices
	   << " " << (*pointIt)[5] + nVertices << ::std::endl;

      file << " " << (*pointIt)[0] + nVertices
	   << " " << (*pointIt)[1] + nVertices
	   << " " << (*pointIt)[6] + nVertices << ::std::endl;

      file << " " << (*pointIt)[1] + nVertices
	   << " " << (*pointIt)[2] + nVertices
	   << " " << (*pointIt)[6] + nVertices << ::std::endl;

      file << " " << elementIt->vertexInfo[2]->outputIndex 
	   << " " << (*pointIt)[1] + nVertices
	   << " " << (*pointIt)[2] + nVertices << ::std::endl;
     
      
    }

  }


  void VtkWriter::writeConnectivity_dim2_degree4(::std::ofstream &file)
  {
    ::std::list<ElementInfo> *elements = (*dc_)[0]->getElementInfos();
    ::std::list<ElementInfo>::iterator elementIt;

    ::std::vector< ::std::vector<int> > *interpPoints = (*dc_)[0]->getInterpPoints();
    ::std::vector< ::std::vector<int> >::iterator pointIt;

    int nVertices = (*dc_)[0]->getNumberVertices();

    for (pointIt = interpPoints->begin(), elementIt = elements->begin(); 
	 pointIt != interpPoints->end(); 
	 ++pointIt, ++elementIt) {

      file << " " << elementIt->vertexInfo[0]->outputIndex 
	   << " " << (*pointIt)[5] + nVertices
	   << " " << (*pointIt)[6] + nVertices << ::std::endl;

      file << " " << (*pointIt)[5] + nVertices
	   << " " << (*pointIt)[9] + nVertices
	   << " " << (*pointIt)[6] + nVertices << ::std::endl;

      file << " " << (*pointIt)[6] + nVertices
	   << " " << (*pointIt)[7] + nVertices
	   << " " << (*pointIt)[9] + nVertices << ::std::endl;

      file << " " << (*pointIt)[7] + nVertices
	   << " " << (*pointIt)[9] + nVertices
	   << " " << (*pointIt)[10] + nVertices << ::std::endl;

      file << " " << (*pointIt)[7] + nVertices
	   << " " << (*pointIt)[8] + nVertices
	   << " " << (*pointIt)[10] + nVertices << ::std::endl;

      file << " " << (*pointIt)[0] + nVertices
	   << " " << (*pointIt)[8] + nVertices
	   << " " << (*pointIt)[10] + nVertices << ::std::endl;

      file << " " << elementIt->vertexInfo[1]->outputIndex 
	   << " " << (*pointIt)[0] + nVertices
	   << " " << (*pointIt)[8] + nVertices << ::std::endl;

      file << " " << (*pointIt)[4] + nVertices
	   << " " << (*pointIt)[5] + nVertices
	   << " " << (*pointIt)[9] + nVertices << ::std::endl;

      file << " " << (*pointIt)[4] + nVertices
	   << " " << (*pointIt)[9] + nVertices
	   << " " << (*pointIt)[11] + nVertices << ::std::endl;

      file << " " << (*pointIt)[9] + nVertices
	   << " " << (*pointIt)[10] + nVertices
	   << " " << (*pointIt)[11] + nVertices << ::std::endl;

      file << " " << (*pointIt)[1] + nVertices
	   << " " << (*pointIt)[10] + nVertices
	   << " " << (*pointIt)[11] + nVertices << ::std::endl;

      file << " " << (*pointIt)[0] + nVertices
	   << " " << (*pointIt)[1] + nVertices
	   << " " << (*pointIt)[10] + nVertices << ::std::endl;

      file << " " << (*pointIt)[3] + nVertices
	   << " " << (*pointIt)[4] + nVertices
	   << " " << (*pointIt)[11] + nVertices << ::std::endl;

      file << " " << (*pointIt)[2] + nVertices
	   << " " << (*pointIt)[3] + nVertices
	   << " " << (*pointIt)[11] + nVertices << ::std::endl;

      file << " " << (*pointIt)[1] + nVertices
	   << " " << (*pointIt)[2] + nVertices
	   << " " << (*pointIt)[11] + nVertices << ::std::endl;

      file << " " << elementIt->vertexInfo[2]->outputIndex 
	   << " " << (*pointIt)[2] + nVertices
	   << " " << (*pointIt)[3] + nVertices << ::std::endl;

    }
  }

  int VtkWriter::updateAnimationFile(::std::string valueFilename,
				     ::std::vector< ::std::string > *paraViewAnimationFrames,
				     const char *animationFilename)
  {
    size_t found = valueFilename.find_last_of("/\\");
    paraViewAnimationFrames->push_back(valueFilename.substr(found + 1));

    ::std::ofstream file;
    file.open(animationFilename);

    file << "<?xml version=\"1.0\"?>" << ::std::endl;
    file << "<VTKFile type=\"Collection\" version=\"0.1\" >"  << ::std::endl;
    file << "<Collection>" << ::std::endl;

    int counter = 0;
    ::std::vector< ::std::string >::iterator it;
    for (it = paraViewAnimationFrames->begin(); 
	 it < paraViewAnimationFrames->end(); 
	 ++it, counter++) {
      file << "<DataSet timestep=\"" << counter
	   << "\" part=\"0\" file=\"" << (*it) << "\"/>" << ::std::endl;      
    }

    file << "</Collection>" << ::std::endl;
    file << "</VTKFile>" << ::std::endl;

    file.close();
    
    return 0;
  }


}
